==> netsa-io-silk_2.12-1.5.3-sources.jar: io/ZlibOutputStreamBuffer.scala <==
// Copyright 2015-2022 by Carnegie Mellon University
// See license information in LICENSE.txt

package org.cert.netsa.io.silk
package io

import java.io.{DataOutputStream, OutputStream}
import java.util.zip.Deflater


/**
  * A [[BufferWriter]] which compresses data using zlib and writes it
  * to a stream in a form compatible with SiLK file streams, using
  * Hadoop compression codecs.
  *
  * @see [[ZlibOutputStreamBuffer$ the companion object]] for more
  * details.
  */
private[silk] class ZlibOutputStreamBuffer(
  outStream:     OutputStream)
    extends BufferWriter
{
  // The output stream
  private val out: DataOutputStream = outStream match {
    case out: DataOutputStream => out
    case _ => new DataOutputStream(outStream)
  }

  // The compression engine
  private val compressor: Deflater = new Deflater(Deflater.DEFAULT_COMPRESSION, false)

  /**
    * Return the size of a buffer that is guaranteed to be large
    * enough to hold the result of compressing `sourceLen` bytes of
    * data.
    */
  private[this] def compressBound(sourceLen: Int): Int = {
    // Formula taken from body of the compressBound function in the
    // zlib 1.2.11 source code.
    sourceLen + (sourceLen >>> 12) + (sourceLen >>> 14) + (sourceLen >>> 25) + 13
  }

  /**
    * Compress and write any pending data, finalize the compression
    * engine, and flush the output stream.
    */
  def end(): Unit = {
    assert(compressor.needsInput)
    //println("end called")
    compressor.end()
    out.flush()
  }

  /**
    * Write the first `length` bytes of data in `buffer` to the output
    * stream.
    *
    * @throws java.io.IOException when an error occurs.
    */
  def putBuffer(uncompressedData: Array[Byte], uncompressedSize: Int): Unit = {
    if (0 == uncompressedSize) {
      return
    }

    // create buffer to hold the result.  FIXME: Consider moving this
    // array outside the function
    val maxCompressedSize = compressBound(uncompressedSize)
    val compressedData = Array.ofDim[Byte](maxCompressedSize)

    //println("putBuffer called with buffer size = " + uncompressedSize +
    //  "; maxCompressedSize = " + maxCompressedSize)

    // compress it
    assert(compressor.needsInput)
    compressor.setInput(uncompressedData, 0, uncompressedSize)
    compressor.finish()

    val compressedSize = compressor.deflate(compressedData, 0, maxCompressedSize)

    //println("Actual compressed size is " + compressedSize)
    //println("total uncompressed bytes: " + compressor.getBytesRead() +
    //  "; total compressed bytes: " + compressor.getBytesWritten())
    //println("Compressor needsInput is " + compressor.needsInput +
    //  " and finished is " + compressor.finished())

    // write the two sizes and the compressed block
    out.writeInt(compressor.getBytesWritten().toInt)
    out.writeInt(uncompressedSize)
    out.write(compressedData, 0, compressedSize)

    // the following block should not be necessary but it is here
    // "just in case"
    while ( !compressor.finished() ) {
      val size = compressor.deflate(compressedData, 0, maxCompressedSize)
      //println("Writing an additional block of " + size + " bytes")
      out.write(compressedData, 0, size)
    }

    assert(compressor.needsInput)
    assert(0 == compressor.deflate(compressedData, 0, maxCompressedSize))
    compressor.reset()
  }
}


/**
  * The ZlibOutputStreamBuffer object provides support for creating an
  * [[ZlibOutputStreamBuffer]].
  */
private[silk] object ZlibOutputStreamBuffer {
  def apply(outStream: OutputStream): ZlibOutputStreamBuffer = {
    new ZlibOutputStreamBuffer(outStream)
  }
}

// @LICENSE_FOOTER@
//
// Copyright 2015-2022 Carnegie Mellon University. All Rights Reserved.
//
// This material is based upon work funded and supported by the
// Department of Defense and Department of Homeland Security under
// Contract No. FA8702-15-D-0002 with Carnegie Mellon University for the
// operation of the Software Engineering Institute, a federally funded
// research and development center sponsored by the United States
// Department of Defense. The U.S. Government has license rights in this
// software pursuant to DFARS 252.227.7014.
//
// NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING
// INSTITUTE MATERIAL IS FURNISHED ON AN "AS-IS" BASIS. CARNEGIE MELLON
// UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR
// IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF
// FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS
// OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT
// MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT,
// TRADEMARK, OR COPYRIGHT INFRINGEMENT.
//
// Released under a GNU GPL 2.0-style license, please see LICENSE.txt or
// contact permission@sei.cmu.edu for full terms.
//
// [DISTRIBUTION STATEMENT A] This material has been approved for public
// release and unlimited distribution. Please see Copyright notice for
// non-US Government use and distribution.
//
// Carnegie Mellon(R) and CERT(R) are registered in the U.S. Patent and
// Trademark Office by Carnegie Mellon University.
//
// This software includes and/or makes use of third party software each
// subject to its own license as detailed in LICENSE-thirdparty.tx
//
// DM20-1143

==> netsa-io-silk_2.12-1.6.0-sources.jar: io/ZlibOutputStreamBuffer.scala <==
// Copyright 2015-2022 by Carnegie Mellon University
// See license information in LICENSE.txt

package org.cert.netsa.io.silk
package io

import java.io.{DataOutputStream, OutputStream}
import java.util.zip.Deflater


/** A [[BufferWriter]] which compresses data using zlib and writes it to
  * a stream in a form compatible with SiLK file streams.
  *
  * @see [[ZlibOutputStreamBuffer$ the companion object]] for more
  * details.
  */
private[silk] class ZlibOutputStreamBuffer(outStream: OutputStream)
    extends BufferWriter
{
  // The output stream
  private val out: DataOutputStream = outStream match {
    case out: DataOutputStream => out
    case _ => new DataOutputStream(outStream)
  }

  // The compression engine
  private val deflater: Deflater =
    new Deflater(Deflater.DEFAULT_COMPRESSION, false)

  /** Return the size of a buffer that is guaranteed to be large enough
    * to hold the result of compressing `sourceLen` bytes of data.
    */
  private def compressBound(sourceLen: Int): Int = {
    // Formula taken from body of the compressBound function in the
    // zlib 1.2.11 source code.
    sourceLen + (sourceLen >>> 12) + (sourceLen >>> 14) + (sourceLen >>> 25) + 13
  }

  /** Compress and write any pending data, finalize the compression
    * engine, and flush the output stream.
    */
  def end(): Unit = {
    assert(deflater.needsInput)
    deflater.end()
    out.flush()
  }

  /** Write the first `length` bytes of data in `buffer` to the output
    * stream.
    *
    * @throws java.io.IOException when an error occurs.
    */
  def putBuffer(uncompressedData: Array[Byte], uncompressedSize: Int): Unit =
    if ( uncompressedSize > 0 ) {

      // create buffer to hold the result.  FIXME: Consider moving this
      // array outside the function
      val maxCompressedSize = compressBound(uncompressedSize)
      val compressedData = Array.ofDim[Byte](maxCompressedSize)

      // compress it
      assert(deflater.needsInput())
      deflater.setInput(uncompressedData, 0, uncompressedSize)
      deflater.finish()

      val compressedSize =
        deflater.deflate(compressedData, 0, maxCompressedSize)

      // write the two sizes and the compressed block
      out.writeInt(compressedSize)
      out.writeInt(uncompressedSize)
      out.write(compressedData, 0, compressedSize)

      assert(deflater.finished())
      assert(deflater.needsInput())
      deflater.reset()
    }
}


/**
  * The ZlibOutputStreamBuffer object provides support for creating an
  * [[ZlibOutputStreamBuffer]].
  */
private[silk] object ZlibOutputStreamBuffer {
  def apply(outStream: OutputStream): ZlibOutputStreamBuffer = {
    new ZlibOutputStreamBuffer(outStream)
  }
}

// @LICENSE_FOOTER@
//
// Copyright 2015-2022 Carnegie Mellon University. All Rights Reserved.
//
// This material is based upon work funded and supported by the
// Department of Defense and Department of Homeland Security under
// Contract No. FA8702-15-D-0002 with Carnegie Mellon University for the
// operation of the Software Engineering Institute, a federally funded
// research and development center sponsored by the United States
// Department of Defense. The U.S. Government has license rights in this
// software pursuant to DFARS 252.227.7014.
//
// NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING
// INSTITUTE MATERIAL IS FURNISHED ON AN "AS-IS" BASIS. CARNEGIE MELLON
// UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR
// IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF
// FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS
// OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT
// MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT,
// TRADEMARK, OR COPYRIGHT INFRINGEMENT.
//
// Released under a GNU GPL 2.0-style license, please see LICENSE.txt or
// contact permission@sei.cmu.edu for full terms.
//
// [DISTRIBUTION STATEMENT A] This material has been approved for public
// release and unlimited distribution. Please see Copyright notice for
// non-US Government use and distribution.
//
// Carnegie Mellon(R) and CERT(R) are registered in the U.S. Patent and
// Trademark Office by Carnegie Mellon University.
//
// This software includes and/or makes use of third party software each
// subject to its own license as detailed in LICENSE-thirdparty.tx
//
// DM20-1143
